#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <time.h>

#include <mpi.h>

#include "func.h"

// programma dimostrativo deve essere migliorato se non altro
// nell'aspetto e nella gestione degli errori

int main (int argc, char ** argv) 
{
  int size, rank, rlen, n = -1, i, j;
  char myname[MPI_MAX_PROCESSOR_NAME]; 
  double ** a, ** b, ** c, * temp, time, 
         mtime, ctime, commtime;

  MPI_Init (&argc, &argv);
  MPI_Comm_size (MPI_COMM_WORLD, &size);
  MPI_Comm_rank (MPI_COMM_WORLD, &rank);
  MPI_Get_processor_name (myname, &rlen);

  fprintf (stdout, "I am proces %3d of %3d my name is %s (%s)\n", 
    rank, size, myname, argv[0]);
  fflush(stdout);

  if (argc == 2)
    n = atoi(argv[1]);

  if (n < 0)
    n = 100;

  if (rank == 0) {
    if ((a = (double **) malloc (sizeof (double *) * n)) == NULL) {
      fprintf (stderr, "Memory allocation error\n");
      MPI_Abort (MPI_COMM_WORLD, 1);
      return EXIT_FAILURE;
    }
    else {
      if ( (temp = (double *) malloc (sizeof (double) * n * n)) == NULL) {
        fprintf (stderr, "Memory allocation error\n");
        MPI_Abort (MPI_COMM_WORLD, 1);
        return EXIT_FAILURE;
      }
      else {
        for (i=0; i<n; i++) 
          a[i] = &temp[i*n]; 
      }
    }

    if ((b = (double **) malloc (sizeof (double *) * n)) == NULL) {
      fprintf (stderr, "Memory allocation error\n");
      MPI_Abort (MPI_COMM_WORLD, 1);
      return EXIT_FAILURE;
    }
    else {
      if ( (temp = (double *) malloc (sizeof (double) * n * n)) == NULL) {
        fprintf (stderr, "Memory allocation error\n");
        MPI_Abort (MPI_COMM_WORLD, 1);
        return EXIT_FAILURE;
      }
      else {
        for (i=0; i<n; i++) 
          b[i] = &temp[i*n]; 
      }
    }

    if ((c = (double **) malloc (sizeof (double *) * n)) == NULL) {
      fprintf (stderr, "Memory allocation error\n");
      MPI_Abort (MPI_COMM_WORLD, 1);
      return EXIT_FAILURE;
    }
    else {
      if ( (temp = (double *) malloc (sizeof (double) * n * n)) == NULL) {
        fprintf (stderr, "Memory allocation error\n");
        MPI_Abort (MPI_COMM_WORLD, 1);
        return EXIT_FAILURE;
      }
      else {
        for (i=0; i<n; i++) 
          c[i] = &temp[i*n]; 
      }
    }

    srand (10);
    for (i=0; i<n; i++) {
      for (j=0; j<n; j++) {
        a[i][j] = 1.0e0 + (double) 
          (1.0e0 * n * (rand() / (RAND_MAX + 1.0e0)));
        b[i][j] = 1.0e0 + (double)
          (1.0e0 * n * (rand() / (RAND_MAX + 1.0e0)));
        c[i][j] = 0.0e0;
      }
    }
  }
  else {
    if ((b = (double **) malloc (sizeof (double *) * n)) == NULL) {
      fprintf (stderr, "Memory allocation error\n");
      MPI_Abort (MPI_COMM_WORLD, 1);
      return EXIT_FAILURE;
    }
    else {
      if ( (temp = (double *) malloc (sizeof (double) * n * n)) == NULL) {
        fprintf (stderr, "Memory allocation error\n");
        MPI_Abort (MPI_COMM_WORLD, 1);
        return EXIT_FAILURE;
      }
      else {
        for (i=0; i<n; i++) 
          b[i] = &temp[i*n]; 
      }
    }
  }

  MPI_Barrier (MPI_COMM_WORLD);
  time = MPI_Wtime ();
  parallel (rank, size, n, a, b, c, &ctime, 
      &commtime);
  MPI_Barrier (MPI_COMM_WORLD);
  time = MPI_Wtime () - time;

  MPI_Reduce (&time, &mtime, 1, MPI_DOUBLE, MPI_MAX, 0, 
      MPI_COMM_WORLD);
  MPI_Reduce (&commtime, &time, 1, MPI_DOUBLE, MPI_MAX, 0, 
      MPI_COMM_WORLD);
  commtime = time;
  MPI_Reduce (&ctime, &time, 1, MPI_DOUBLE, MPI_SUM, 0, 
      MPI_COMM_WORLD);
  ctime = time;

  if (rank == 0)
    fprintf (stdout, "Parallel time: Tot %f s, Comp %f s, Comm %f s\n", 
        mtime, ctime, commtime);

  if (rank == 0) {
    double ** c1;

    if ((c1 = (double **) malloc (sizeof (double *) * n)) == NULL) {
      fprintf (stderr, "Memory allocation error\n");
      MPI_Abort (MPI_COMM_WORLD, 1);
      return EXIT_FAILURE;
    }
    else {
      if ( (temp = (double *) malloc (sizeof (double) * n * n)) == NULL) {
        fprintf (stderr, "Memory allocation error\n");
        MPI_Abort (MPI_COMM_WORLD, 1);
        return EXIT_FAILURE;
      }
      else {
        for (i=0; i<n; i++) 
          c1[i] = &temp[i*n]; 
      }
    }

    for (i=0; i<n; i++) 
      for (j=0; j<n; j++)
        c1[i][j] = 0.0e0;

    time = (double) clock();
    sequential (n, a, b, c1);
    time = (double) clock() - time;

    fprintf (stdout, "Sequential time: %f \n", time/CLOCKS_PER_SEC);

    for (i=0; i<n; i++) {
      for (j=0; j<n; j++) {
        if (fabs(c[i][j]-c1[i][j]) > 1.0e-5) {
          fprintf (stdout, "Error at %d:%d (%f)\n", i, j, c[i][j] - c1[i][j]);
        }
      }
    }

    /*
    for (i=0; i<n; i++) {
      for (j=0; j<n; j++) {
        fprintf (stdout, "%f ", c[i][j]);
      }
      fprintf (stdout, "\n");
    }

    fprintf (stdout, "\n");

    for (i=0; i<n; i++) {
      for (j=0; j<n; j++) {
        fprintf (stdout, "%f ", c1[i][j]);
      }
      fprintf (stdout, "\n");
    }
    */

    free (c1[0]);
    free (c1);

    free (a[0]);
    free (c[0]);
    free (a);
    free (c);
  }

  free (b[0]);
  free (b);

  MPI_Finalize ();

  return EXIT_SUCCESS;
}
