#include <stdio.h>
#include <stdlib.h>

#include <mpi.h>

#include "func.h"

void parallel (int rank, int size, int n, double ** a, 
    double ** b, double ** c, double * ctime,
    double * commtime) 
{
  int i, j, k, dim;
  double * rows, * result, time;

  dim = n/size;

  if ( (rows = (double *) malloc (sizeof (double) * dim
          * n)) == NULL) {
    fprintf (stderr, "Memory allocation error\n");
    MPI_Abort (MPI_COMM_WORLD, 1);
    exit (EXIT_FAILURE);
  }

  if ( (result = (double *) malloc (sizeof (double) * dim *
          n)) == NULL) {
    fprintf (stderr, "Memory allocation error\n");
    MPI_Abort (MPI_COMM_WORLD, 1);
    exit (EXIT_FAILURE);
  }

  for (i=0; i<dim; i++)
    for (j=0; j<n; j++)
      result[(i*n) + j] = 0.0e0;

  time = MPI_Wtime ();
  MPI_Scatter (a[0], n*dim, MPI_DOUBLE, rows, n*dim, MPI_DOUBLE, 
      0, MPI_COMM_WORLD); 
  MPI_Bcast (b[0], n*n, MPI_DOUBLE, 0, MPI_COMM_WORLD);
  time = MPI_Wtime() - time;
  *commtime = time; 

  time = MPI_Wtime ();
  for (i=0; i<dim; i++)
    for (k=0; k<n; k++)
      for (j=0; j<n; j++)
        result[(i*n) + j] = result[(i*n) + j] + 
          rows[(i*n) + k] * b[k][j];
  time = MPI_Wtime() - time;
  *ctime = time;

  time = MPI_Wtime ();
  MPI_Gather (result, dim*n, MPI_DOUBLE, c[0], dim*n, MPI_DOUBLE, 
      0, MPI_COMM_WORLD );
  time = MPI_Wtime() - time;

  *commtime += time;

  if (rank == 0) {
    if ((dim*size) != n) {
      time = MPI_Wtime ();
      for (i=(dim*size); i<n; i++)
        for (k=0; k<n; k++)
          for (j=0; j<n; j++)
            c[i][j] = c[i][j] + a[i][k] * b[k][j];
      time = MPI_Wtime() - time;

      *ctime += time;
    }
  }

  free (rows);
  free (result);

  return;
}
